#
# shinGETsu - P2P anonymous BBS
#
# (c) FukTommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: Gateway.pm,v 1.32 2005/01/09 15:56:20 fuktommy Exp $
#

package Shingetsu::Gateway;
use Shingetsu::AttachCache;
use Shingetsu::Config;
use Shingetsu::Message;
use Shingetsu::Signature;
use Shingetsu::Util;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(str_encode str_decode file_encode file_decode rec escape xdie
		args argsFromMulti touch readNote post newElement
		addRecord margeRecord search postDeleteMessage
		bracket_link print302 print404 print403 printHeader checkAdmin
		html_format xlocaltime readQuery
		deleteDialog deleteRecord deleteFile deleteFileDialog
		deleteRecordDialog deleteFileForm deleteRecordForm
		checkSign newElementForm
);

#
# configuration
#
$datadir	= $Shingetsu::Config::datadir;
$cachedir	= $Shingetsu::Config::cachedir;
$root		= $Shingetsu::Config::root;
$message	= new Shingetsu::Message();

#
# tips
#
sub str_encode($) {		# Thanks for YukiWiki & WalWiki
	my($s) = @_;
	$s =~ s|[^\w]|'%' . uc(unpack('H2', $&))|eg;
	return $s;
}
sub str_decode($) {		# Thanks for YukiWiki
	my ($s) = @_;
	$s =~ s/\+/ /g;
	$s =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C", hex($1))/eg;
	return escape($s);
}
sub file_encode($$) {		# Thanks for YukiWiki
	my ($type, $name) = @_;
	$name =~ s/./uc(unpack('H2', $&))/eg;
	return "${type}_${name}";
}
sub file_decode($) {		# Thanks for YukiWiki
	my($name) = @_;
	@_ = split /_/, $name, 2;
	$name = $_[1];
	$name =~ s/[0-9A-F][0-9A-F]/pack("C", hex($&))/eg;
	return escape($name);
}
sub rec($) {
	return Shingetsu::Util::rec($_[0]);
}
sub xdie(;@) {
	print "Content-Type: text/plain;\r\n",
	      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
	      "\r\n",
	      @_, "\n";
	exit;
}

#
# escape
#
sub escape($) {
	local($_) = @_;
	s/&/&amp;/g;
	s/&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);/&$1;/g;
	s/</&lt;/g;
	s/>/&gt;/g;
	s/\r//g;
	s/\n/<br>/g;
	return $_;
}

#
# unescape arguments
#
sub args($) {
	local $_ = $_[0];
	local @_ = split /&/;
	my %arg = ();
	foreach (@_) {
		local @_ = split /=/;
		if (defined $_[1]) {
			$_[1] =~ s/\+/ /g;
			$_[1] =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
			$arg{$_[0]} = escape($_[1]);
		} else {
			$arg{$_[0]} = "";
		}
	}
	return %arg;
}

#
# unescape arguments(multipart/form-data)
#
sub argsFromMulti($) {
	my($input) = @_;
	local $_;
	my %arg = ();
	$input =~ /\s/;
	my $boundary = $`;
	my @input = split $boundary, $input;
	shift @input;
	foreach (@input) {
		my @buf = split /\r\n/, $_;
		next if (@buf == 1);
		shift @buf;
		$buf[0] =~ /Content-Disposition: form-data; name="([^"]+)"/i or next;
		my $key = $1;

		if ($buf[0] =~ /filename="([^"]+)"/i) {
			$arg{auto_suffix} = $1;
			$arg{auto_suffix} =~ s/.*[\/\\]//;
			if ($arg{auto_suffix} =~ /\.([^.]*)$/) {
				$arg{auto_suffix} = $1;
				$arg{auto_suffix} = lc $arg{auto_suffix};
			} else {
				$arg{auto_suffix} = "";
			}
		}

		$_ = shift @buf until ((! defined $_) || ($_ eq ""));
		$arg{$key} = join "\r\n", @buf;
		if ($key ne "attach") {
			$arg{$key} = escape($arg{$key});
		}
	}
	return %arg;
}

#
# date
#
sub xlocaltime(;$) {
	my @buf = Shingetsu::Util::localtime(@_);
	return sprintf "%d-%02d-%02d %02d:%02d",
		       1900+$buf[5], $buf[4]+1, $buf[3], $buf[2], $buf[1];
}

#
# make file
#
sub touch($) {
	my($file) = @_;
	if ($ENV{HTTP_USER_AGENT} =~ /$Shingetsu::Config::robot/) {
		return;
	}
	my %stat = Shingetsu::CacheStat::list();
	unless (-s "$datadir/$file.dat") {
		Shingetsu::Util::lock();
		open OUT, ">> $datadir/$file.dat";
		close OUT;
		Shingetsu::Util::unlock();
		my %stat = Shingetsu::CacheStat::list();
		$stat{$file}{records} = 0;
		$stat{$file}{stamp}   = 0;
		$stat{$file}{size}    = 0;
		Shingetsu::CacheStat::sync(%stat);
	}
	if ((-s "$datadir/$file.dat") == 0) {
		my $node = Shingetsu::Cache::search($file);
		if (defined $node) {
			Shingetsu::Cache::getRegion($file, $node);
		}
	}
}

#
# read note history
#
sub readNote($) {
	my($file) = @_;
	local $_;
	my $buf = "";
	my %ref = ();
	if (open IN, "$datadir/$file.dat") {
		while (<IN>) {
			chomp;
			my %rec = rec($_);
			next if ((keys %rec <= 2) || (defined $rec{remove_stamp}));
			$buf = $_;
			$ref{"$rec{stamp}_$rec{id}"} = 0 if ($rec{base_stamp});
			if ($rec{base_stamp} && (defined $ref{"$rec{base_stamp}_$rec{base_id}"})) {
				$ref{"$rec{base_stamp}_$rec{base_id}"}++;
			}
		}
		close IN;
	}
	my $orphan = grep $ref{$_}==0, keys %ref;
	$orphan = 0 if ($orphan == 1);
	return ($buf, $orphan);
}

#
# add record
#
sub addRecord($$$) {
	my($file, $stamp, $body) = @_;
	my $id = Shingetsu::Extern::md5digest($body);
	Shingetsu::Util::lock();
	open OUT, ">> $datadir/$file.dat" or die "$datadir/$file.dat: failed to open. $!";
	print OUT "$stamp<>$id<>$body\n";
	close OUT;
	Shingetsu::Util::unlock();
	my %stat = Shingetsu::CacheStat::list();
	$stat{$file}{records}++;
	$stat{$file}{stamp} = $stamp;
	$stat{$file}{size}  = (-s "$datadir/$file.dat");
	Shingetsu::CacheStat::sync(%stat);
	Shingetsu::AttachCache::addRecord($file, "$stamp<>$id<>$body");
	return $id;
}

#
# post article
#
sub post($%) {
	my($thread, %arg) = @_;
	local $_;
	my ($suffix, $attach) = ("", "");
	$arg{name} = "" unless (defined $arg{name});
	$arg{mail} = "" unless (defined $arg{mail});
	$arg{passwd} = "" unless (defined $arg{passwd});
	if ((defined $arg{attach}) && ($arg{attach} ne "")) {
		if (length $arg{attach} > $Shingetsu::Config::filelimit*1024*1024) {
			xdie $$message{big_file};
		}
		$attach = Shingetsu::Extern::base64encode($arg{attach});
		if ($arg{suffix} ne "AUTO") {
			$suffix	= $arg{suffix};
		} elsif ((defined $arg{auto_suffix}) && ($arg{auto_suffix} ne "")) {
			$suffix	= $arg{auto_suffix};
		} else {
			$suffix	= "txt";
		}
	}
	my $stamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body  = "";
	my $target = "";
	$arg{suffix} = $suffix;
	$arg{attach} = $attach;
	foreach ("base_stamp", "base_id","name", "mail", "body", "suffix", "attach") {
		if ((defined $arg{$_}) && ($arg{$_} ne "")) {
			$body .= "<>$_:$arg{$_}";
			$target .= ",$_";
		}
	}
	$body =~ s/^<>//;
	$target =~ s/^,//;
	xdie $$message{null_article} if ($body eq "");
	if ($arg{passwd} ne "") {
		my($pubkey, $sign) = Shingetsu::Signature::generate($arg{passwd}, $body);
		$body = "pubkey:$pubkey<>sign:$sign<>target:$target<>" . $body;
	}
	my $id = addRecord($thread, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $thread);
		Shingetsu::Cache::tellupdate($thread, $stamp, $id, undef);
	}
	return substr $id, 0, 8;
}

#
# make new list element
#
sub newElement($%) {
	my($file, %arg) = @_;
	local $_;
	my $stamp = Shingetsu::Util::time();
	xdie $$message{null_name} if ($arg{name} eq "");
	xdie $$message{bad_name}  if ($arg{name} =~ m/\/|\]\]/);
	xdie $$message{null_type} if ($arg{type} eq "");
	my $body = "type:$arg{type}<>link:$arg{name}";
	my $id = addRecord($file, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $file);
		Shingetsu::Cache::tellupdate($file, $stamp, $id, undef);
	}
}

#
# add new record and marge that and last record (for note)
#
sub margeRecord {
	my($file, %arg) = @_;
	my $stamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body = $arg{message};
	my $base_stamp = "";
	my $base_id    = "";
	local $_;
	my($buf, $conflict) = readNote($file);
	if ($buf) {
		my %rec = rec($buf);
		if ($rec{body} ne "") {
			if ($rec{body} =~ /<br>$/) {
				$body = "$rec{body}<br>$body";
			} else {
				$body = "$rec{body}<br><br>$body";
			}
		}
		$base_stamp = $rec{stamp};
		$base_id    = $rec{id};
	}
	xdie $$message{null_article} if ($body eq "");
	$body = "body:$body";
	$body .= "<>base_stamp:$base_stamp<>base_id:$base_id" if ($base_stamp);
	my $id = addRecord($file, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $file);
		Shingetsu::Cache::tellupdate($file, $stamp, $id, undef);
	}
	return substr $id, 0, 8;
}

#
# search string
# return: @filename
#
sub search($$) {
	my($type, $query) = @_;
	my $file;
	my @result = ();
	foreach $file (glob("$datadir/${type}_*.dat")) {
		my $file2 = $file;
		$file2 =~ s|^$datadir/||; $file2 =~ s|.dat$||;
		my $file3  = "$cachedir/$file2/body.cache";
		my $x = file_decode($file2);
		my $y = str_encode($x);
		my $found = 0;
		if ($x =~ /\Q$query\E/i) {
			$found = 1;
		} else {
			open IN, $file3 or open IN, $file
				or die "$file: open failed. $!";
			while (<IN>) {
				if (/\Q$query\E/i) {
					$found = 1;
					last;
				}
			}
			close IN;
		}
		if ($found) {
			push @result, $file2;
		}
	}
	return @result;
}

#
# post delete message to other nodes
#
sub postDeleteMessage($$$%) {
	my($file, $stamp, $id, %arg) = @_;
	$arg{name} = "" unless (defined $arg{name});
	$arg{mail} = "" unless (defined $arg{mail});
	$arg{passwd} = "" unless (defined $arg{passwd});
	my $newstamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body  = "";
	my $target = "";
	if ((defined $arg{name})&&($arg{name} ne "")) {
		$body .= "<>name:$arg{name}";
		$target .= ",name";
	}
#	if ((defined $arg{mail})&&($arg{mail} ne "")) {
#		$body .= "<>mail:$arg{mail}";
#		$target .= ",mail";
#	}
	if ((defined $arg{message})&&($arg{message} ne "")) {
		$body .= "<>body:$arg{message}";
		$target .= ",body";
	}
	$body .= "<>remove_stamp:$stamp";	$target .= ",remove_stamp";
	$body .= "<>remove_id:$id";		$target .= ",remove_id";
	$body =~ s/^<>//;
	$target =~ s/^,//;
	xdie $$message{null_article} if ($body eq "");
	if ($arg{passwd} ne "") {
		my($pubkey, $sign) = Shingetsu::Signature::generate($arg{passwd}, $body);
		$body = "pubkey:$pubkey<>sign:$sign<>target:$target<>" . $body ;
	}
	my $newid = addRecord($file, $newstamp, $body);
	Shingetsu::Util::addUpdate($newstamp, $newid, $file);
	Shingetsu::Cache::tellupdate($file, $newstamp, $newid, undef);
}

#
# encode bracket string to link
#
sub bracket_link($) {
	my($str) = @_;
	my $link = $str;

	if ($link =~ m{^/(thread)/([^/]+)/([0-9a-f]{8})$}) {
		$link = "${root}$1.cgi/" . str_encode($2) . "#r$3";
		return "<a href='$link'>[[$str]]</a>";

	} elsif ($link =~ m{^/(list|thread|note)/([^/]+)$}) {
		$link = "${root}$1.cgi/" . str_encode($2);
		return "<a href='$link'>[[$str]]</a>";

	} elsif ($link =~ m{^([^/]+)/([0-9a-f]{8})$}) {
		$link = str_encode($1) . "#r$2";
		return "<a href='$link'>[[$str]]</a>";

	} elsif ($link =~ m{^([^/]+)$}) {
		$link = str_encode($1);
		return "<a href='$link'>[[$str]]</a>";

	} else {
		return "[[$str]]";
	}



}

#
# check admin address
#
sub checkAdmin() {
	if ($ENV{REMOTE_ADDR} !~ $Shingetsu::Config::adminAddr) {
		print403();
		exit;
	}
}

#
# print CGI&HTML header
# $opt{rss}
#
sub printHeader($%) {
	my($title, %opt) = @_;
	my $ver = $Shingetsu::Config::version;
	my $rss	= $opt{rss} || "${root}gateway.cgi/rss";
	print "Content-Type: text/html; charset=UTF-8\r\n",
	      "Content-Language: $$message{lang}\r\n",
	      "X-Shingetsu: shinGETsu/$ver\r\n",
	      "\r\n";
	print "<?xml version='1.0' encoding='UTF-8'?>\n",
	      "<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.1//EN'\n",
	      "  'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd'>\n",
	      "<html xmlns='http://www.w3.org/1999/xhtml' xml:lang='$$message{lang}'>\n",
	      "<head>\n",
	      "  <title>$title</title>\n",
	      "  <link rev='made' href='http://shingetsu.sourceforge.net/' />\n",
	      "  <link rel='contents' href='${root}list.cgi' />\n",
	      "  <link rel='stylesheet' type='text/css' href='$Shingetsu::Config::css' />\n",
	      "  <link rel='alternate' type='application/rss+xml' title='RSS' href='$rss' />\n",
	      "</head>\n";
	print "<body>\n",
	      "<p class='head'>\n",
	      "  <a href='${root}gateway.cgi'>$$message{top}</a> |\n",
	      "  <a href='${root}list.cgi'>$$message{menu}</a> |\n",
	      "  <a href='${root}gateway.cgi/new'>$$message{new}</a> |\n",
	      "  <a href='${root}gateway.cgi/index'>$$message{index}</a> |\n",
	      "  <a href='${root}gateway.cgi/changes'>$$message{changes}</a> |\n",
	      "  <a href='${root}gateway.cgi/update'>$$message{update}</a> |\n",
	      "  <a href='${root}gateway.cgi/search'>$$message{search}</a> |\n",
	      "  <a href='$rss'>RSS</a>\n",
	      "</p>\n",
	      "<h1><a href='${root}gateway.cgi/search/", str_encode($title),
	      "' class='title' title='$$message{search}'>$title</a></h1>\n"
}

#
# print CGI header (403 forbidden)
#
sub print403(;$) {
	my ($file) = @_;
	printHeader($$message{403});
	print "</body></html>\n";
}

#
# print CGI header (404 not found)
#
sub print404(;$) {
	my ($file) = @_;
	printHeader($$message{404});
	print "<p>$$message{try_later}</p>";
	if (defined $file) {
	      print "<form method='get' action='$ENV{SCRIPT_NAME}/delete?file=$file'>",
		    "<p><input type='submit' value='$$message{del_file}' tabindex='1' accesskey='d' /></p></form>";
	}
	print "</body></html>\n";
}

#
# print CGI header (302 moved temporarily)
#
sub print302($) {
	my($next) = @_;
	my $ver = $Shingetsu::Config::version;
	my($title) = @_;
	print "Content-Type: text/html; charset=UTF-8\r\n",
	      "Content-Language: en\r\n",
	      "X-Shingetsu: shinGETsu/$ver\r\n",
	      "\r\n",
	      '<?xml version="1.0" encoding="UTF-8"?>',
	      '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"',
	      ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
	      '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">',
	      '<head>',
	      "<title>moved</title>",
	      "<link rev='made' href='http://shingetsu.sourceforge.net/' />",
	      "<link rel='contents' href='${root}gateway.cgi/list' />",
	      "</head>",
	      "<body><p>Click and jump to <a href='$next'>$next</a></p>\n",
	      "<script type='text/javascript'>window.location.href = '$next';</script>",
	      "</body></html>\n";
}

#
# format to HTML
#
sub html_format($) {
	local($_) = @_;
	s|<br>|<br />\n    |g;
	s|(&gt;&gt;)([0-9A-Za-z]+)|<a href='#r$2'>$&</a>|g;
	s|https?://[^\x00-\x20"'()<>\[\]\x7F-\xFF]{2,}|<a href='$&'>$&</a>|g;
	s|\[\[<a.*?>(.*?)\]\]</a>|[[$1]]|g;
	s|\[\[([^<>]+?)\]\]|bracket_link($1)|eg;
	if ($Shingetsu::Config::redirect) {
		s|<a href='http://shingetsu\.p2p/|<a href='/|g;
	}
	return $_;
}

#
# read query
#
sub readQuery() {
	my($path, $input);
	if ($ENV{REMOTE_ADDR} !~ $Shingetsu::Config::friendAddr) {
		print403();
		exit;
	} elsif (defined $ENV{PATH_INFO}) {
		$path = $ENV{PATH_INFO};
		$path =~ s|^/||;
	} else {
		$path = "";
	}

	if ($ENV{REQUEST_METHOD} eq "POST") {
		read(STDIN, $input, $ENV{CONTENT_LENGTH});
	} else {
		$input = $ENV{QUERY_STRING};
	}
	return ($path, $input);
}

#
# delete record form
#
sub deleteRecordForm($$$$$) {
	my($file, $buf, $stamp, $id, $arg) = @_;
	$buf =~ s/</&lt;/g;
	$buf =~ s/>/&gt;/g;
	print	"<p>$$message{del_record_q}</p>",
		"<form method='post' action='$ENV{SCRIPT_NAME}'><p>",
		"<input type='hidden' value='xdelete' name='cmd' />",
		"<input type='hidden' value='$file' name='file' />",
		"<input type='hidden' value='$stamp' name='stamp' />",
		"<input type='hidden' value='$id' name='id' />",
		"<input type='submit' value='$$message{yes}' name='submit' tabindex='1' accesskey='w' />",
		" <input type='checkbox' value='dopost' name='dopost' tabindex='2' accesskey='s' />",
		"$$message{send}",
		" <input type='checkbox' value='error' name='error'",
		" checked='checked' tabindex='2' accesskey='e' />",
		"$$message{error}<br />",
		" $$message{name}:<input name='name' size='15' value='' tabindex='3' accesskey='n' />",
		" $$message{signature}:<input type='password' name='passwd' size='15' value='' tabindex='4' accesskey='p' />",
		" $$message{comment}:<input name='message' size='15' value='' tabindex='5' accesskey='m' />",
		"<input type='hidden' name='mode' value='$$arg{mode}' />",
		"</p></form>\n",
		"<p>$buf</p>";
}

#
# delete file form
#
sub deleteFileForm($@) {
	my($file, @buf) = @_;
	print "<p>$$message{del_file_q} ",
	      "<a href='$ENV{SCRIPT_NAME}?cmd=xdelete&amp;file=$file'>",
	      "$$message{yes}</a></p>\n";
	if ($file =~ /_/) {
		print "<p>", file_decode($file), "($$message{$`})</p>\n";
	}
	foreach (@buf) {
		chomp;
		s/</&lt;/g;
		s/>/&gt;/g;
		print "<p>$_</p>\n";
	}
}

#
# delete record dialog
#
sub deleteRecordDialog(%) {
	my %arg = @_;
	printHeader($$message{del_record});
	my $file = $arg{file};
	print "\n";
	my($stamp, $id) = split /\//, $arg{record};
	my $buf;
	if (open IN, "$cachedir/$file/body.cache") {
		while (<IN>) {
			if (/^$stamp<>$id<>/) {
				$buf = $_;
				last;
			}
		}
		close IN;
		if (! defined $buf) {
			print "<p>$$message{no_record}</p></body></html>\n";
		} else {
			deleteRecordForm($file, $buf, $stamp, $id, \%arg);
			print "</body></html>\n";
		}
	} else {
		print "<p>$file: $$message{no_file}</p></body></html>\n";
	}
}

#
# delete file dialog
#
sub deleteFileDialog(%) {
	my %arg = @_;
	my $file = $arg{file};
	printHeader("$$message{del_file}");
	print "\n";
	my @buf = ();
	my $count = 0;
	if ((open IN, "$cachedir/$file/body.cache")
	   ||(open IN, "$datadir/$file.dat")) {
		while (<IN>) {
			push @buf, $_;
			last if ($count++ > 1);
		}
		close IN;
		deleteFileForm($file, @buf);
		print "</body></html>\n";
	} else {
		print "<p>$file: $$message{no_file}</p></body></html>\n";
	}
}

#
# delete file
#
sub deleteFile {
	my($file) = @_;
	if (Shingetsu::Cache::removeFile($file)) {
		print302("${root}list.cgi");
	} else {
		print "Content-Type: text/plain\r\n\r\n",
		      "failed. $!\n";
	}
}

#
# delete record
#
sub deleteRecord($$$$) {
	my($file, $stamp, $id, $arg) = @_;
	my %arg = %$arg;
	local $_;
	if (! Shingetsu::Cache::removeRecord($file, $stamp, $id)) {
		print "Content-Type: text/plain\r\n\r\n",
		      "failed. $!\n";
		return;
	}
	if ($arg{dopost}) {
		postDeleteMessage($file, $stamp, $id, %arg);
	}
	if ((defined $arg{mode}) && ($arg{mode} ne "")) {
		my $xfile = str_encode(file_decode($file));
		print302("$ENV{SCRIPT_NAME}/$xfile");
	}
}

#
# delete dialog
#
sub deleteDialog(%) {
	my %arg = @_;
	local $_;
	if (! defined $arg{file}) {
		print404();
		exit;
	} elsif (defined $arg{record}) {
		deleteRecordDialog(%arg);
	} else {
		deleteFileDialog(%arg);
	}
}

#
# check signature
#
sub checkSign(%) {
	my %arg = @_;
	unless ($arg{file} && $arg{stamp} && $arg{id}) {
		print404();
		exit;
	}
	my $file = $arg{file};
	touch($file);
	if (open IN, "$cachedir/$file/record/$arg{stamp}_$arg{id}.cache") {
		my $value;
		local $_ = <IN>;
		close IN;
		my %rec = rec($_);
		if (! ((defined $rec{pubkey}) && ($rec{pubkey} ne "")
			&& (defined $rec{sign}) && ($rec{sign} ne "")
			&& (defined $rec{target}) && ($rec{target} ne ""))) {
			print "Content-Type: text/plain; charset=UTF-8\r\n",
			      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
			      "\r\n",
			      "$$message{not_signed}\n$_\n";
		} elsif (Shingetsu::Signature::check(%rec)) {
			s/</&lt;/g; s/>/&gt;/g;
			my $trip = Shingetsu::Signature::pubkey2trip($rec{pubkey});
			my($type, $filename) = split /_/, $file, 2;
			printHeader($$message{check_sign});
			print "<p>$$message{protected}($rec{target}).</p>",
			"<form method='get' action='${root}gateway.cgi/edittrust'><p>",
			"<input type='submit' name='button' value='$$message{add}' tabindex='1' accesskey='a' />",
			"/<input type='submit' name ='button' value='$$message{delete}' tabindex='2' accesskey='d' />",
			"<input type='hidden' name='trip' value='$trip' />",
			"<input type='hidden' name='type' value='$type' />",
			"<input type='hidden' name='file' value='$filename' />",
			"$$message{to_from_trust_list}: $trip</p></form>",
			"<p>$_</p>",
			"</body></html>";
			close IN;
			exit;
		} else {
			print "Content-Type: text/plain; charset=UTF-8\r\n",
			      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
			      "\r\n",
			      "$$message{wrong_sign}.\n$_\n";
			close IN;
			exit;
		}
	}
	close IN;
	print404($file);
}

#
# new element form
# $opt{file}, $opt{checked}, $opt{dopost}
#
sub newElementForm(%) {
	my %opt = @_;
	$opt{checked} = "" unless ($opt{checked});
	my $file = (defined $opt{file})? "$opt{file}": "";

	print "<form method='get' action='$ENV{SCRIPT_NAME}'><p>\n",
	      "  <input type='hidden' name='cmd' value='new' />\n",
	      "  <input type='hidden' name='file' value='$file' />\n",
	      "  <input type='submit' value='$$message{new}' name='submit' tabindex='1' accesskey='w' />\n",
	      "  <input type='radio' value='list' name='type' tabindex='2' accesskey='t' />",
	      "$$message{list}\n",
	      "  <input type='radio' value='thread' name='type' tabindex='2' accesskey='t' />",
	      "$$message{thread}\n",
	      "  <input type='radio' value='note' name='type' tabindex='2' accesskey='t' />",
	      "$$message{note}\n";
	print "  <input type='checkbox' value='dopost' name='dopost'",
	      " $opt{checked} tabindex='2' accesskey='s' />",
	      "$$message{send}\n" if ($opt{dopost});
	print "  <br />\n",
	      "  $$message{name}: <input name='name' size='19' value='' tabindex='3' accesskey='t' />\n",
	      "</p></form>\n";
}

1;
