#!/usr/bin/perl -w
#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: httpd.pl,v 1.55 2004/12/19 11:40:37 fuktommy Exp $
#

use strict;
use lib qw(.);
use Socket;
use IPC::Open2;
use Shingetsu::Config;
use Shingetsu::Extern;
use Shingetsu::Util;

use vars qw( $debug $pid_client $timeout $bufsize);

sub xdie(@)   { killClient(); die @_; }
sub dprint(@) { print STDOUT @_, "\n" if ($debug); }
sub xlocaltime(;$) {
	my @buf = Shingetsu::Util::localtime(@_);
	return sprintf "%d-%02d-%02d %02d:%02d:%02d",
		       1900+$buf[5], $buf[4]+1, $buf[3], $buf[2], $buf[1], $buf[0];
}
sub killClient {
	if ((defined $pid_client)&&($pid_client>0)) {
		kill "TERM", $pid_client;
		dprint "terminate $pid_client";
	}
	unlink $Shingetsu::Config::pid;
}
sub str_decode($) {		# from YukiWiki
	my ($s) = @_;
	$s =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C", hex($1))/eg;
	return $s;
}

#
# print HTTP header
#
sub printHeader($) {
	my ($stat) = @_;
	my $ver = $Shingetsu::Config::version;
	alarm $timeout;
	print "HTTP/1.0 $stat\r\n",
	      "Server: shinGETsu/$ver\r\n",
	      "Connection: Close\r\n",
	      "Date: ", Shingetsu::Util::rfc822_time(), "\r\n",
	      "";
	alarm 0;
}

#
# print HTTP header and body for error
#
sub printError($) {
	my($stat) = @_;
	printHeader($stat);
	alarm $timeout;
	print "Content-Type: text/plain\r\n",
	      "\r\n",
	      "$stat\n";
	alarm 0;
}

#
# print cgi result
# $opt{range}, $opt{gzip}
#
sub printResult($%) {
	my($pbuf, %opt) = @_;
	my @buf = @$pbuf;
	my @head = ();
	my $range = $opt{range};
	while ($buf[0] !~ /^\s*$/) {
		push @head, shift(@buf);
	}
	shift @buf;
	my $buf = join "", @buf;
	my $len = length $buf;

	if ($range) {
		my $all = length $buf;
		$buf = substr $buf, $range;
		$len = length $buf;
		printHeader("206 Partial content");
		print "Content-Range: bytes $range-" . ($all-1)  . "/$all\r\n",
		      "Content-Length: $len\r\n";

	} elsif ($opt{gzip}) {
		my $buf_gzip = Shingetsu::Extern::xgzip($buf);
		if (length($buf_gzip) < length($buf)) {
			$buf = $buf_gzip;
			$len = length $buf;
			printHeader("200 OK");
			print "Content-Encoding: gzip\r\n",
			      "Content-Length: $len\r\n";
		} else {
			printHeader("200 OK");
			print "Content-Length: $len\r\n";
		}

	} else {
		printHeader("200 OK");
		print "Content-Length: $len\r\n";
	}
	print @head;
	print "\r\n";

	if ($ENV{REQUEST_METHOD} ne "HEAD") {
		my $i;
		alarm $timeout;
		for ($i=0; $i<$len; $i+=$bufsize) {
			print substr $buf, $i, $bufsize;
			alarm $timeout;
		}
		alarm 0;
	}
}

#
# print file content
# $opt{range}
#
sub printFile($%) {
	my($file, %opt) = @_;
	$file =~ s|^/+||;
	$file =~ s|/+$||;
	$file =~ s|^\.\.+||;
	$file =~ s|/\.\./|/|g;
	my $filedir = $Shingetsu::Config::wwwdir;
	if (grep $_ eq $file, @Shingetsu::Config::systemFile) {
		$filedir = $Shingetsu::Config::filedir;
	}

	my $encoding;
	if (-d "$filedir/$file") {
		local $_;
		foreach (@Shingetsu::Config::index) {
			if (-f "$filedir/$file/$_") {
				$file = "$file/$_";
				last;
			}
		}
	}
	$file =~ /\.([^.]+)$/;
	my $suffix = lc $1;
	my $base   = $`;
	if ($Shingetsu::Config::encoding{$suffix}) {
		$encoding = $Shingetsu::Config::encoding{$suffix};
		$base =~ /\.([^.]+)$/;
		$suffix = lc $1;
	}
	$suffix = "txt" unless (defined $suffix);
	my $type = $Shingetsu::Config::mimeType{$suffix};
	$type = "text/plain" unless (defined $type);
	my $range = $opt{range};
	if ((-f "$filedir/$file")&&(open IN, "$filedir/$file")) {
		my @stat = stat "$filedir/$file";
		printHeader("200 OK");
		alarm $timeout;
		print "Content-Type: $type\r\n",
		      "Last-Modified: ", Shingetsu::Util::rfc822_time($stat[9]), "\r\n";
		print "Content-Encoding: $encoding\r\n" if (defined $encoding);
		if ($range) {
			my $all = (-s "$filedir/$file");
			my $len = $all - $range;
			seek IN, $range, 00;
			print "Content-Range: bytes $range-" . ($all-1)  . "/$all\r\n",
			      "Content-Length: $len\r\n",
			      "\r\n";
		} else {
			print "Content-Length: ", (-s "$filedir/$file"), "\r\n",
			      "\r\n";
		}
		my $buf;
		if ($ENV{REQUEST_METHOD} ne "HEAD") {
			while (read IN, $buf, $bufsize) {
				print $buf;
				alarm $timeout;
			}
		}
		alarm 0;
	} else {
		printError("404 Not Found");
	}
}

#
# exec CGI
#
sub execCGI($) {
	my($cgiDir) = @_;
	$ENV{SCRIPT_FILENAME} = "$cgiDir$ENV{SCRIPT_NAME}";
	my($IN, $OUT);
	my @buf = ();
	if (! (-x ".$ENV{SCRIPT_NAME}")) {
		printError("403 Forbidden");
		return 0;
	} elsif (! open2($IN, $OUT, ".$ENV{SCRIPT_NAME}")) {
		printError("500 Internal Server Error");
		return 0;
	} else {
		if ($ENV{CONTENT_LENGTH}) {
			my $buf;
			alarm $timeout;
			my $i;
			for ($i=0; $i<int($ENV{CONTENT_LENGTH}/$bufsize); $i++) {
				read CLIENT, $buf, $bufsize;
				print $OUT $buf;
				alarm $timeout;
			}
			if ($ENV{CONTENT_LENGTH} % $bufsize) {
				read CLIENT, $buf, $ENV{CONTENT_LENGTH} % $bufsize;
				print $OUT $buf;
			}
			alarm 0;
		}
		close $OUT;
		my @buf = <$IN>;
		close $IN;
		return (1, @buf);
	}
}

#
# httpd
# args: port, cgi-directory-path
#
sub httpd($$) {
	my($port, $cgiDir) = @_;

	# initialize daemon
	socket CLIENT_WAITING, PF_INET, SOCK_STREAM, 0 or xdie "failed to open socket. $!\n";
	setsockopt CLIENT_WAITING, SOL_SOCKET, SO_REUSEADDR, 1 or xdie "failed to setsockopt. $!\n";
	bind CLIENT_WAITING, pack_sockaddr_in($port, INADDR_ANY) or xdie "failed to bind. $!\n";
	listen CLIENT_WAITING, SOMAXCONN or xdie "failed to listen: $!\n";

	# run daemon
	while (1) {
		my($paddr);
		sleep 1 if (rand() < 1/$Shingetsu::Config::childrenLimit);
		unless ($paddr = accept(CLIENT, CLIENT_WAITING)) {
			#dprint "faild to accept";
			sleep 1;
			next;
		}
		my ($client_port, $client_iaddr) = unpack_sockaddr_in $paddr;
		my $client_ip = inet_ntoa $client_iaddr;
		my $client_hostname;
		if ($Shingetsu::Config::resolv) {
			$client_hostname = gethostbyaddr($client_iaddr, AF_INET) || $client_ip;
		} else {
			$client_hostname = $client_ip;
		}

		select CLIENT; $|=1;
		my $pid;

		if (! defined($pid = fork)) {
			dprint "server process: fork failed.";
			close CLIENT;
			next;

		} elsif ($pid) {
			close CLIENT;
			next;

		} else {
			$SIG{TERM} = $SIG{HUP} = $SIG{CHLD} = "DEFAULT";
			$SIG{ALRM} = sub { die "$client_ip: server timeout" };
			$ENV{REMOTE_HOST} = $client_hostname;
			$ENV{REMOTE_ADDR} = $client_ip;
			$ENV{REMOTE_PORT} = $client_port;

			# read HTTP request
			my($str_request, $referer, $agent) = ("", "", "");
			my($gzip, $range);
			alarm $timeout;
			while (<CLIENT>) {
				s/[\n|\r]//g;
				if ($_ eq "") {
					last;

				} elsif (/^(\S+)\s+(\S+)/ && (!defined $ENV{REQUEST_METHOD})) {
					$str_request = $_;
					$ENV{REQUEST_METHOD} = $1;
					$ENV{REQUEST_URI} = $2;
					$ENV{REQUEST_URI} =~ s|^//+|/|g;

				} elsif (/^Accept-Encoding:.*gzip/i) {
					$gzip = 1;

				} elsif (/^Range:\s*bytes\s*=\s*(\d+)\s*-/) {
					$range = $1;

				} elsif (/^Content-length:\s+(\d+)/i) {
					$ENV{CONTENT_LENGTH} = $1;

				} elsif (/^Content-Type:\s+([^;]+)/i) {
					$ENV{CONTENT_TYPE} = $1;
				}
				if (/^([^:]+):\s+/) {
					my $key = uc $1;
					my $val = $';
					$key =~ s/-/_/g;
					$ENV{"HTTP_$key"} = $val;
				}
			}
			alarm 0;
			my $buf = "";
			my $flag_error;
			if ((defined $ENV{CONTENT_LENGTH})
				&& ($ENV{CONTENT_LENGTH} > $Shingetsu::Config::filelimit*1024*1024)) {
				$flag_error = 1;
			}

			if ($ENV{REQUEST_URI} =~ m|^[a-z]+://([^/]+)|) {
				$ENV{HTTP_HOST}   = $1;
				$ENV{REQUEST_URI} = $';
			}
			if ($ENV{REQUEST_URI} eq "") {
				$ENV{REQUEST_URI} = "/";
			}

			if (($ENV{REQUEST_URI} =~ m|^/([0-9a-zA-Z]+)\.cgi|i)
				&& (-f "$1.cgi")) {
				$ENV{SCRIPT_NAME}  = "/$1.cgi";
				$ENV{PATH_INFO}	   = $';
				if ($ENV{PATH_INFO} =~ /\?/) {
					$ENV{QUERY_STRING} = $';
					$ENV{PATH_INFO}	  = $`;
				} else {
					$ENV{QUERY_STRING} = "";
				}
				$ENV{PATH_INFO} = str_decode($ENV{PATH_INFO});

			} elsif ($ENV{REQUEST_URI} eq "/") {
				$ENV{SCRIPT_NAME}  = $Shingetsu::Config::index;
				$ENV{QUERY_STRING} = "";

			}

			# output logs
			{	local $^W = 0;
				my $date = xlocaltime;
				dprint "$client_ip<>$date<>$str_request<>$ENV{HTTP_REFERER}<>$ENV{HTTP_USER_AGENT}";

			}

			$gzip = 0 if ($ENV{SCRIPT_NAME}
				&& ($ENV{SCRIPT_NAME} ne "/server.cgi")
				&& ($ENV{SCRIPT_NAME} ne "/hosting.cgi"));

			if ($flag_error) {
				printError("404 Not Found");

			} elsif ($ENV{REQUEST_METHOD} eq "OPTIONS") {
				printHeader("200 OK");
				print "Allow: GET, HEAD, POST, OPTIONS\n",
				      "Content-Length: 0\n\n";

			} elsif ($ENV{REQUEST_METHOD} !~ /^(GET|HEAD|POST)$/) {
				printHeader("501 Not Implemented");

			} elsif (defined $ENV{SCRIPT_NAME}) {
				my($ok, @buf) = execCGI($cgiDir);
				if ($ok) {
					printResult(\@buf, range=>$range, gzip=>$gzip);
				}

			} else {
				printFile($ENV{REQUEST_URI}, range=>$range);
			}

			shutdown CLIENT, 2;
			close CLIENT;
			wait; # for CGI
			exit;
		}
	}
}

#
# client process
#
sub client {
	$SIG{TERM} = $SIG{HUP} = "DEFAULT";
	$SIG{CHLD} = "IGNORE";
	sleep 5;
	while (1) {
		my $pid;
		if (! defined ($pid = fork)) {
			dprint "client process: fork failed.";
			sleep $Shingetsu::Config::clientfrec * 60;
			next;

		} elsif ($pid) {
			sleep 10;
			my $t = $Shingetsu::Config::clientfrec * 60
			      + int($Shingetsu::Config::timeError * Shingetsu::Util::normal());
			$t = $Shingetsu::Config::clientfrec * 60 if ($t <= 0);
			sleep $t;
			# wait; # ?????
			next;

		} else {
			open STDOUT, "> /dev/null";
			$ENV{QUERY_STRING}   = "";
			$ENV{PATH_INFO}	     = "";
			$ENV{REQUEST_METHOD} = "GET";
			$ENV{REMOTE_ADDR}    = "127.0.0.1";
			$ENV{SCRIPT_NAME}    = "/client.cgi";
			$SIG{CHLD} = "DEFAULT";
			exec "./client.cgi";
		}
	}
}

#------------------------------------------------------------------#
$SIG{CHLD} = "IGNORE";
$SIG{TERM} = $SIG{HUP} = sub {
	killClient();
	exit;
};

#
# default value
#
my $port	 = $Shingetsu::Config::port;
my $cgiDir	 = $Shingetsu::Config::cgiDir;
$debug		 = $Shingetsu::Config::debug;
$timeout	 = $Shingetsu::Config::timeout;
$bufsize	 = $Shingetsu::Config::bufsize;
my @initNode = ();

#
# arguments
#
while (@ARGV) {
	$_ = shift @ARGV;
	if ($_ eq "-v") {
		$debug = 1;
	} elsif ($_ eq "-p") {
		$port = shift @ARGV;
	} elsif ($_ eq "-i") {
		push @initNode, shift(@ARGV);
	}
}
chdir $cgiDir or die "$cgiDir: no such directory. $!\n";
foreach ($Shingetsu::Config::datadir, $Shingetsu::Config::cachedir) {
	mkdir $_, 0777;
}
unlink $Shingetsu::Config::client;
unless ($debug) {
	my $pid;
	if ($pid = fork) {
		$SIG{TERM} = $SIG{HUP} = "DEFAULT";
		open OUT, "> $Shingetsu::Config::pid";
		print OUT "$pid\n";
		close OUT;
		exit;
	}
	open STDIN,  "< /dev/null";
	open STDOUT, "> /dev/null";
	open STDERR, "> /dev/null";
}

#
# environment variable
#
$ENV{SERVER_SOFTWARE}	= "shinGETsu/$Shingetsu::Config::version";
$ENV{GATEWAY_INTERFACE} = "CGI/1.1";
$ENV{SERVER_PROTOCOL}	= "HTTP/1.0";
$ENV{SERVER_PORT}	= $port;
$ENV{HTTP_PORT}		= $port;
$ENV{SERVER_NAME}	= "";
$ENV{HTTP_HOST}		= "";
$ENV{X_INIT_NODE}	= join "+", @initNode if (@initNode);

#
# client process
#
if (!($pid_client = fork)) {
	client();
}

#
# daemon process
#
httpd($port, $cgiDir);
